//==- CodingConventionBestPractices.cpp - assert good code -------*- C++ -*-==//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file defines the checks for "best practices" in the context of coding
// conventions.
//
//===----------------------------------------------------------------------===//

#include "clang/Analysis/CodingConventions.h"

#include "CodingConventionChecker.h"
#include "clang/AST/Decl.h"
#include "clang/AST/DeclVisitor.h"
#include "clang/AST/Stmt.h"
#include "clang/AST/StmtVisitor.h"
#include "clang/Basic/SourceManager.h"
#include "clang/Lex/Preprocessor.h"
#include <sstream>
#include <fstream>
#include <cerrno>
#include <map>

using namespace clang;


namespace {


//===----------------------------------------------------------------------===//
// Code Checker.
//===----------------------------------------------------------------------===//
  class RuleChecker: public CodingConventionConfigReader {
  CodingConventionChecker& CCC;
  static std::string RuleFileName;

public:
  RuleChecker(CodingConventionChecker& ccc) :
    CCC(ccc) { }
  bool processConfigLine(llvm::StringRef Line, size_t LineNo);
  void checkFunctionSignature(FunctionDecl* D);
  void checkPostUnaryOp(UnaryOperator* S);
};


//===----------------------------------------------------------------------===//
// Code Checker Stmt Visitor.
//===----------------------------------------------------------------------===//
  class WalkAST: public DeclVisitor<WalkAST>,
                 public StmtVisitor<WalkAST> {
  RuleChecker &Checker;
  typedef DeclVisitor<WalkAST> DeclVisitorBase;
  typedef StmtVisitor<WalkAST> StmtVisitorBase;

public:
  WalkAST(RuleChecker& checker) :
    Checker(checker)
  {}

  void Visit(Decl* S) { DeclVisitorBase::Visit(S); }
  void Visit(Stmt* S) { StmtVisitorBase::Visit(S); }

  void VisitStmt(Stmt* S) { VisitChildren(S); }
  void VisitDecl(Decl* D) { VisitChildren(D); }

  void VisitChildren(Stmt* S) {
    for (Stmt::child_iterator i = S->child_begin(), e = S->child_end();
         i != e; ++i)
      Visit(*i);
  }

  void VisitChildren(Decl *D) {
    DeclContext* C = dyn_cast<DeclContext>(D);
    if (C) {
      for (DeclContext::decl_iterator i = C->decls_begin(), e = C->decls_end();
           i != e; ++i)
        Visit(*i);
    }
    if (isa<TemplateDecl>(D)) {
      Decl* ND = dyn_cast<TemplateDecl>(D)->getTemplatedDecl();
      if (ND) {
        Visit(ND);
      }
    } else if (isa<VarDecl>(D)) {
      Stmt* VI = dyn_cast<VarDecl>(D)->getInit();
      if (VI) {
        Visit(VI);
      }
    }
  }

  void Treat(Decl* D) {}

  void Treat(FunctionDecl* D) {
    Checker.checkFunctionSignature(dyn_cast<FunctionDecl>(D));
  }

#define DECL(Derived, Base) \
  void Visit##Derived##Decl(Derived##Decl * D) { \
    Stmt* S = D->getBody(); if (S) Visit(S); \
    Treat(D); \
    VisitChildren(D); \
  }
#define ABSTRACT_DECL(Derived, Base) DECL(Derived, Base)
#include "clang/AST/DeclNodes.def"

  void VisitDeclStmt(DeclStmt* S) {
    if (S->isSingleDecl()) {
      Decl* D = S->getSingleDecl();
      Stmt* B = D->getBody();
      if (B) Visit(B);
    } else {
      VisitChildren(S);
    }
  }

  void VisitCompoundStmt(CompoundStmt* S) {
    for (CompoundStmt::body_iterator i = S->body_begin(), e = S->body_end();
         i != e; ++i)
      Visit(*i);
    VisitChildren(S);
  }

  void VisitUnaryOperator(UnaryOperator* E) {
    if (E->getOpcode() == UnaryOperator::PostInc
        || E->getOpcode() == UnaryOperator::PostDec) {
      Checker.checkPostUnaryOp(E);
    VisitChildren(E);
    }
  }

};

} // end anonymous namespace


void RuleChecker::checkPostUnaryOp(UnaryOperator* E) {
  SourceLocation L = E->getOperatorLoc();
  if (!L.isValid() || !L.isFileID()) return;

  if (E->getOpcode() == UnaryOperator::PostInc) {
    CCC.report(E, "use ++a instead of a++.", CodeModificationHint());
  } else {
    CCC.report(E, "use --a instead of a--.", CodeModificationHint());
  }
}

void RuleChecker::checkFunctionSignature(FunctionDecl* D) {
  if (!isa<FunctionType>(D->getType().getTypePtr()))
    return;

  if (!D->hasWrittenPrototype())
    return;

  unsigned NumParams = D->getNumParams();
  if (NumParams == 0) return;

  ASTContext& AST = D->getASTContext();
  uint64_t PtrSize = AST.getTypeSize(AST.VoidPtrTy);
  for (unsigned i = 0; i != NumParams; ++i) {
    ParmVarDecl* P = D->getParamDecl(i);
    QualType QT = P->getType();
    Type* T = QT.getTypePtr();
    if (T->isAnyPointerType() || T->isReferenceType())
      continue;
    if (AST.getTypeSize(T) > PtrSize) {
      PrintingPolicy PP(CCC.getLangOptions());
      PP.SuppressTagKind = true;
      PP.SuppressScope = true;
      PP.ConstantArraySizeAsWritten = true;
      std::string Str("better pass const ");
      Str += QT.getAsString(PP);
      Str += "& instead of copying the parameter";
      CCC.report(P, Str, CodeModificationHint());
    }
  }

}

bool RuleChecker::processConfigLine(llvm::StringRef Line, size_t LineNo) {
  return true;
}

void clang::CheckCodingConventionBestPractices(Decl *D, Diagnostic& Diags,
                                               SourceManager& SM,
                                               const LangOptions& Lang) {
  CodingConventionChecker CCC(Diags, SM, Lang);
  RuleChecker Checker(CCC);
  CCC.readConfig("BestPractice", Checker);
  WalkAST walker(Checker);
  walker.Visit(D);
}

