//==- CodingConvention.cpp - Check for name rules ------*- C++ -*-==//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file defines the checks for identifiers in the context of coding
// convention checks.
//
//===----------------------------------------------------------------------===//

#include "clang/Analysis/CodingConventions.h"

#include "CodingConventionChecker.h"
#include "clang/Basic/Diagnostic.h"
#include "clang/Basic/SourceManager.h"
#include "clang/Frontend/FrontendDiagnostic.h"
#include <sstream>
#include <fstream>
#include <cerrno>
#include <map>

using namespace clang;


void clang::SetCodingConventionConfigFile(llvm::StringRef FileName) {
  CodingConventionChecker::setConfigFile(FileName);
}

std::string clang::CodingConventionChecker::ConfigFileName = "coderules.txt";

void clang::CodingConventionChecker::setConfigFile(llvm::StringRef filename) {
  ConfigFileName = filename;
}

void clang::CodingConventionChecker::report(const SourceLocation& Start,
                                            const SourceRange& Range,
                                            llvm::StringRef Str,
                                            const CodeModificationHint& Fixit) {
  unsigned DiagID = Diags.getCustomDiagID(Diagnostic::Error, Str);

  CodeModificationHint CMH;
  Diags.Report(FullSourceLoc(Start, SM), DiagID) << Range << Fixit;
}

bool clang::CodingConventionChecker::readConfig(llvm::StringRef Section,
                                             CodingConventionConfigReader& CR) {
  std::ifstream In(ConfigFileName.c_str());
  if (!In) {
    std::string Msg("Cannot read file: ");
    Msg += strerror(errno);
    reportConfigError(Msg, (size_t)-1);
    return false;
  }
  std::string Line;
  size_t LineNo = 0;
  bool SectionFound = false;

  while (std::getline(In, Line)) {
    if (!In) break;
    ++LineNo;

    size_t posSectionTag = 0;
    while (isspace(Line[posSectionTag]))
      ++posSectionTag;
    SectionFound = Line[posSectionTag] == '['
      && Line[posSectionTag + Section.size() + 1] == ']'
      && Line.compare(posSectionTag + 1, Section.size(), Section) == 0;
    if (SectionFound) break;
  }

  if (!SectionFound) return false;

  while (std::getline(In, Line)) {
    if (!In) break;
    ++LineNo;

    // Skip space-only / empty lines.
    size_t firstNonSpace = 0;
    while (Line[firstNonSpace]) {
      if (!isspace(Line[firstNonSpace])) break;
      ++firstNonSpace;
    }
    if (!Line[firstNonSpace] || Line[firstNonSpace] == '#')
      continue;
    if (Line[firstNonSpace] == '[') // Next section.
      break;

    if (!CR.processConfigLine(Line, LineNo))
      break;
  }
  return true;
}

void clang::CodingConventionChecker::reportConfigError(llvm::StringRef Str,
                                                       size_t LineNo) {
  std::stringstream SErr;
  SErr << "Error in coding convention config file " << ConfigFileName;
  if (LineNo != (size_t)-1)
    SErr << ", line " << LineNo;
  SErr << ": " << Str.data();
  unsigned DiagID = Diags.getCustomDiagID(Diagnostic::Fatal, SErr.str());
  Diags.Report(DiagID);
}

clang::CodingConventionConfigReader::~CodingConventionConfigReader()
{}
