//==- CodingConventionNames.cpp - Check for name rules -----------*- C++ -*-==//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file defines the checks for identifiers in the context of coding
// convention checks.
//
//===----------------------------------------------------------------------===//

#include "clang/Analysis/CodingConventions.h"

#include "CodingConventionChecker.h"
#include "clang/AST/DeclVisitor.h"
#include "clang/AST/Decl.h"
#include <sstream>
#include <set>
#include <map>
#include <vector>

using namespace clang;


namespace {

//===----------------------------------------------------------------------===//
// An entry in the rule set.
//===----------------------------------------------------------------------===//

class NameRule {
public:
  enum ECheck {
    Nothing,
    UpperCase,
    LowerCase,
    AllCaps,

    StartString,
    EndString,

    NumChecks,

    FirstWithArg = StartString
  };

  enum EDeclKind {
    Decl,
#define DECL(Derived, Base) Derived ## Decl = Decl::Derived,
#include "clang/AST/DeclNodes.def"
#define ABSTRACT_DECL(Derived, Base) Derived ## Decl,
#include "clang/AST/DeclNodes.def"
    NumDeclKinds
  };

  NameRule(EDeclKind forwhat = Decl, ECheck check = Nothing, const char* arg=""):
    ForWhat(forwhat), Check(check), Arg(arg) {}

  void check(clang::NamedDecl* N, CodingConventionChecker& CCC,
             const std::string& callCondition = "") const;

  bool operator<(const NameRule& O) const { return Check < O.Check; }
  bool operator==(const NameRule& O) const { return Check == O.Check; }

  void addConditionalRule(const std::string& cond, NameRule rule) const {
    Conditionals[cond] = rule;
  }

private:
  void report(clang::NamedDecl* N, CodingConventionChecker& CCC,
              const std::string& Condition,
              const char* fixit = 0) const;

  const char* getNameOfDecl(EDeclKind kind) const {
    switch (kind) {
    default: assert(0 && "Declaration not in DeclNodes.def!"); break;
#define DECL(Derived, Base) case Derived ## Decl: return #Derived;
#define ABSTRACT_DECL(Derived, Base) DECL(Derived, Base)
#include "clang/AST/DeclNodes.def"
    }
    return "";
  }

  EDeclKind ForWhat;
  ECheck Check;
  std::string Arg;
  mutable std::map<std::string /*condition*/, NameRule> Conditionals;
}; 

//===----------------------------------------------------------------------===//
// Code Checker.
//===----------------------------------------------------------------------===//

class NameChecker: public CodingConventionConfigReader {
  CodingConventionChecker& CCC;
  std::vector<std::set<NameRule> > RuleSet;

  static const int kNumDeclKinds = NameRule::NumDeclKinds;

public:
  NameChecker(CodingConventionChecker& ccc) : CCC(ccc)
  {
    RuleSet.resize(kNumDeclKinds);
  }

  void ruleCheck(NamedDecl* N, NameRule::EDeclKind kind);
  int getCheckForName(const std::string& name) const {
    if (name == "upper") return NameRule::UpperCase;
    if (name == "lower") return NameRule::LowerCase;
    if (name == "caps") return NameRule::AllCaps;
    if (name == "starts") return NameRule::StartString;
    if (name == "ends") return NameRule::EndString;
    return -1;
  }

private:
  int getDeclKindForName(const std::string& name) const {
#define DECL(Derived, Base) \
    if (name == #Derived) return (int)NameRule::Derived ## Decl;
#define ABSTRACT_DECL(Derived, Base) DECL(Derived, Base)
#include "clang/AST/DeclNodes.def"
    return -1;
  }

  bool processConfigLine(llvm::StringRef Line, size_t LineNo);
  void endOfConfig();
};

//===----------------------------------------------------------------------===//
// Checker Decl Visitor.
//===----------------------------------------------------------------------===//
class WalkAST: public DeclVisitor<WalkAST> {
  NameChecker& NC;

public:
  WalkAST(NameChecker& nc) : NC(nc) {}

  void visitChildren(Decl* D);
  void visitChildren(TemplateDecl* N);

#define DECL(Derived, Base) \
  void Visit##Derived##Decl(Derived##Decl * D) { \
    NamedDecl* N = dyn_cast<NamedDecl>(D); \
    if (N && !N->isImplicit() \
        && N->getDeclName() && N->getDeclName().isIdentifier()) { \
      NC.ruleCheck(N, NameRule::Derived ## Decl); \
    } \
    visitChildren(D); \
  }
#define ABSTRACT_DECL(Derived, Base) DECL(Derived, Base)
#include "clang/AST/DeclNodes.def"
};

} // end anonymous namespace


//===----------------------------------------------------------------------===//
// String asserters.
//===----------------------------------------------------------------------===//

void NameRule::check(clang::NamedDecl *N, CodingConventionChecker& CCC,
                     const std::string& callCondition) const {

  // General rule only applies if no conditional rule us used.
  for (std::map<std::string /*condition*/, NameRule>::const_iterator
         i = Conditionals.begin(), e = Conditionals.end(); i != e; ++i) {
    std::string Condition = i->first;
    bool Test = Condition[0] != '!';
    if (!Test) {
      Condition.erase(0);
    }
    if (Condition == "staticDataMember") {
      clang::VarDecl* VD = dyn_cast<clang::VarDecl>(N);
      if (VD) {
        if (VD->isStaticDataMember() == Test) {
          i->second.check(N, CCC, Condition);
        }
      }
    }
    else if (Condition == "inNamespace") {
      DeclContext* DC = N->getDeclContext();
      if (DC && DC->isNamespace() == Test) {
        i->second.check(N, CCC, Condition);
      }
    }
  }

  std::string Name = N->getNameAsCString();
  switch (Check) {
  case UpperCase:
    if (!isupper(Name[0])) {
      Name[0] = toupper(Name[0]);
      report(N, CCC, callCondition, Name.c_str());
    }
    return;
  case LowerCase:
    if (!islower(Name[0])) {
      Name[0] = tolower(Name[0]);
      report(N, CCC, callCondition, Name.c_str());
    }
    return;
  case AllCaps: 
    for (size_t i = 0, e = Name.length(); i != e; ++i) {
      if (!isupper(Name[i])) {
        std::transform(Name.begin() + i, Name.end(), Name.begin() + i,
                       (int(*)(int)) std::toupper);
        report(N, CCC, callCondition, Name.c_str());
        return;
      }
    }
    return;
  case StartString:
    if (Name.compare(0, Arg.length(), Arg) != 0) {
      Name[0] = toupper(Name[0]);
      Name = Arg + Name;
      report(N, CCC, callCondition, Name.c_str());
    }
    return;
  case EndString:
    if (Name.compare(Name.length() - Arg.length(), Arg.length(), Arg) != 0) {
      Name += Arg;
      report(N, CCC, callCondition, Name.c_str());
    }
    return;
  default:
    break;
  };
}

void NameRule::report(clang::NamedDecl *N, CodingConventionChecker& CCC,
                      const std::string& Condition,
                      const char* fixit) const {
  static const char* Msgs[] = {
    "be anything they want",
    "start with an uppercase letter",
    "start with a lowercase letter",
    "be all uppercase",
    "start with ",
    "end with "
  };
  std::stringstream S;

  S << "Names of";

  std::string DeclType(getNameOfDecl(ForWhat));
  // beautify decltype
  if (DeclType == "CXXRecord") {
    S << " classe"; // "s" gets added below...
  } else {
    size_t lenDeclType = DeclType.length();
    for (size_t i = 0; i < lenDeclType; ++i) {
      if (isupper(DeclType[i])) {
        S << " ";
        S << (char) tolower(DeclType[i]);
      } else {
        S << DeclType[i];
      }
    }
  }

  S << "s must ";
  S << Msgs[Check];
  if (Check >= FirstWithArg) {
    S << Arg;
  }
  if (!Condition.empty()) {
    S << " if " << Condition;
  }
  S << ".";

  SourceLocation Loc(N->getLocStart());
  SourceRange SR(N->getSourceRange());
  SR.setEnd(N->getLocEnd());

  CCC.report(Loc, SR, S.str(),
             CodeModificationHint::CreateReplacement(Loc, fixit));
}

void NameChecker::ruleCheck(NamedDecl* N, NameRule::EDeclKind kind) {
  for (std::set<NameRule>::const_iterator i = RuleSet[kind].begin(),
         e = RuleSet[kind].end(); i != e; ++i) {
    NameRule NR = *i;
    NR.check(N, CCC);
  }
}

void WalkAST::visitChildren(TemplateDecl *N) {
  TemplateParameterList* parms = N->getTemplateParameters();
  if (parms) {
    for (TemplateParameterList::const_iterator i = parms->begin(),
           e = parms->end(); i != e; ++i) {
      Visit(*i);
    }
  }
  NamedDecl* ND = N->getTemplatedDecl();
  if (ND) {
    Visit(ND);
  }
}

void WalkAST::visitChildren(Decl *N) {
  DeclContext* C = dyn_cast<DeclContext>(N);
  if (C) {
    for (DeclContext::decl_iterator i = C->decls_begin(), e = C->decls_end();
         i != e; ++i)
      Visit(*i);
  }
}

bool NameChecker::processConfigLine(llvm::StringRef Line, size_t LineNo) {

  std::string SArg;
  std::string SDeclKind;
  std::string SCheck;

  std::istringstream InS(Line);

  SDeclKind = "<NOT READ>";
  InS >> SDeclKind;

  if (InS) {
    InS >> SCheck;
  }
  if (!InS) {
    if (SDeclKind == "<NOT READ>") {
      CCC.reportConfigError("cannot read rule's type of declaration.", LineNo);
    } else {
      CCC.reportConfigError("cannot read rule's check.", LineNo);
    }
    RuleSet.clear();
    return false;
  }

  size_t posDot = SDeclKind.find('.');
  std::string Condition;
  if (posDot != std::string::npos) {
    Condition = SDeclKind.substr(posDot + 1, -1);
    SDeclKind.erase(posDot, -1);
  }

  int DeclKind = getDeclKindForName(SDeclKind);
  if (DeclKind == -1) {
    std::string Msg(SDeclKind);
    Msg += " is an unknown type of declaration.";
    CCC.reportConfigError(Msg, LineNo);
    RuleSet.clear();
    return false;
  }

  size_t posParen = SCheck.find('(');
  if (posParen != std::string::npos) {
    SArg = SCheck.substr(posParen + 1, SCheck.length() - posParen - 2);
    SCheck.erase(posParen, -1);
  } else {
    SArg.clear();
  }
  int Check = getCheckForName(SCheck);
  if (Check == -1) {
    std::string Msg(SCheck);
    Msg += " is an unknown check.";
    CCC.reportConfigError(Msg, LineNo);
    RuleSet.clear();
    return false;
  }
  if (Check >= (int) NameRule::FirstWithArg) {
    if (posParen == std::string::npos) {
      std::string Msg("cannot read required argument for check ");
      Msg += SCheck;
      CCC.reportConfigError(Msg, LineNo);
      RuleSet.clear();
      return false;
    }
  } else {
    if (!SArg.empty()) {
      std::string Msg(SCheck);
      Msg += " does not require any argument.";
      CCC.reportConfigError(Msg, LineNo);
      RuleSet.clear();
      return false;
    }
  }

  if (posDot != std::string::npos) {
    std::set<NameRule>::iterator IR
      = RuleSet[DeclKind].find(NameRule((NameRule::EDeclKind) DeclKind,
                                        (NameRule::ECheck) Check,
                                        SArg.c_str()));
    if (IR == RuleSet[DeclKind].end()) {
      IR = RuleSet[DeclKind].insert(NameRule((NameRule::EDeclKind) DeclKind,
                                             NameRule::Nothing)).first;
    }
    IR->addConditionalRule(Condition, NameRule((NameRule::EDeclKind) DeclKind,
                                               (NameRule::ECheck) Check,
                                               SArg.c_str()));
  } else {
    RuleSet[DeclKind].insert(NameRule((NameRule::EDeclKind) DeclKind,
                                      (NameRule::ECheck) Check,
                                      SArg.c_str()));
  }
  return true;
}

void NameChecker::endOfConfig() {
  // Inherit base decl's rules.
#define DECL(Derived, Base) \
  for (std::set<NameRule>::const_iterator i = RuleSet[NameRule::Base].begin(), \
       e = RuleSet[NameRule::Base].end(); i != e; ++i) { \
    std::set<NameRule>* NR = &RuleSet[NameRule::Derived ## Decl]; \
    if (NR->find(*i) == NR->end()) \
      NR->insert(*i); \
  }
#define ABSTRACT_DECL(Derived, Base) DECL(Derived, Base)
#include "clang/AST/DeclNodes.def"

}


void clang::CheckCodingConventionNames(Decl *D, Diagnostic& Diag,
                                       SourceManager& SM,
                                       const LangOptions& Lang) {
  CodingConventionChecker CCC(Diag, SM, Lang);
  NameChecker Checker(CCC);
  CCC.readConfig("Names", Checker);
  WalkAST Walker(Checker);
  Walker.Visit(D);
}

