//==- CodingConventionSpaces.cpp - Check for spacing rules -------*- C++ -*-==//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file defines the checks for code white spaces in the context of coding
// convention checks.
//
//===----------------------------------------------------------------------===//

#include "clang/Analysis/CodingConventions.h"

#include "CodingConventionChecker.h"
#include "clang/AST/Decl.h"
#include "clang/AST/DeclVisitor.h"
#include "clang/AST/Stmt.h"
#include "clang/AST/StmtVisitor.h"
#include "clang/Basic/SourceManager.h"
#include "clang/Lex/Preprocessor.h"
#include <sstream>
#include <map>

using namespace clang;


namespace {


//===----------------------------------------------------------------------===//
// Code Checker.
//===----------------------------------------------------------------------===//
  class RuleChecker: public CodingConventionConfigReader {
    CodingConventionChecker& CCC;
  std::map<std::string, int> Rules;

  static std::string RuleFileName;

public:
  RuleChecker(CodingConventionChecker& ccc) :
    CCC(ccc) {}
  void checkBinaryOp(BinaryOperator* E);
  void checkParen(ParenExpr* E);

private:
  const char* getTokenPtr(SourceLocation& L,
                          unsigned& ColNo);

  bool processConfigLine(llvm::StringRef, size_t LineNo);
};


//===----------------------------------------------------------------------===//
// Code Checker Stmt Visitor.
//===----------------------------------------------------------------------===//
  class WalkAST: public DeclVisitor<WalkAST>,
                 public StmtVisitor<WalkAST> {
  RuleChecker &Checker;
  typedef DeclVisitor<WalkAST> DeclVisitorBase;
  typedef StmtVisitor<WalkAST> StmtVisitorBase;

public:
  WalkAST(RuleChecker& checker) :
    Checker(checker)
  {}

  void Visit(Decl* S) { DeclVisitorBase::Visit(S); }
  void Visit(Stmt* S) { StmtVisitorBase::Visit(S); }

  void VisitStmt(Stmt* S) { VisitChildren(S); }
  void VisitDecl(Decl* D) { VisitChildren(D); }

  void VisitChildren(Stmt* S) {
    for (Stmt::child_iterator i = S->child_begin(), e = S->child_end();
         i != e; ++i)
      Visit(*i);
  }
  void VisitChildren(Decl *D) {
    DeclContext* C = dyn_cast<DeclContext>(D);
    if (C) {
      for (DeclContext::decl_iterator i = C->decls_begin(), e = C->decls_end();
           i != e; ++i)
        Visit(*i);
    }
    if (isa<TemplateDecl>(D)) {
      Decl* ND = dyn_cast<TemplateDecl>(D)->getTemplatedDecl();
      if (ND) {
        Visit(ND);
      }
    } else if (isa<VarDecl>(D)) {
      Stmt* VI = dyn_cast<VarDecl>(D)->getInit();
      if (VI) {
        Visit(VI);
      }
    }
  }

#define DECL(Derived, Base) \
  void Visit##Derived##Decl(Derived##Decl * D) { \
    Stmt* S = D->getBody(); if (S) Visit(S); \
    VisitChildren(D); \
  }
#define ABSTRACT_DECL(Derived, Base) DECL(Derived, Base)
#include "clang/AST/DeclNodes.def"

  void VisitDeclStmt(DeclStmt* S) {
    if (S->isSingleDecl()) {
      Stmt* D = S->getSingleDecl()->getBody();
      if (D) Visit(D);
    } else {
      VisitChildren(S);
    }
  }

  void VisitIfStmt(IfStmt* S) {
    Visit(S->getCond());
    Visit(S->getThen());
    Visit(S->getElse());
    VisitChildren(S);
  }

  void VisitBinaryOperator(BinaryOperator* E) {
    Checker.checkBinaryOp(E);
    VisitChildren(E);
  }

  void VisitParenExpr(ParenExpr* E) {
    Checker.checkParen(E);
    VisitChildren(E);
  }

};

} // end anonymous namespace


std::string RuleChecker::RuleFileName = "coderules.txt";

const char* RuleChecker::getTokenPtr(SourceLocation& L,
                                      unsigned& ColNo) {
  std::pair<FileID, unsigned> LocInfo
    = CCC.getSourceManager().getDecomposedLoc(L);
    FileID FID = LocInfo.first;
    unsigned FileOffset = LocInfo.second;
    ColNo = CCC.getSourceManager().getColumnNumber(FID, FileOffset); // 1-based
    // Get information about the buffer it points into.
    std::pair<const char*, const char*> BufferInfo
      = CCC.getSourceManager().getBufferData(FID);
    const char* BufStart = BufferInfo.first;
    return BufStart+FileOffset;
}

void RuleChecker::checkBinaryOp(BinaryOperator* E) {
  SourceLocation L = E->getOperatorLoc();
  if (!L.isValid() || !L.isFileID()) return;

  unsigned ColNo = 0;
  const char* TokPtr = getTokenPtr(L, ColNo);
  const char* TokEnd = TokPtr
    + Lexer::MeasureTokenLength(L, CCC.getSourceManager(),
                                CCC.getLangOptions());

  if (ColNo > 1) {
    if (!isspace(*(TokPtr-1)) || !isspace(*(TokEnd))) {
      CCC.report(E, "Binary operator must be surrounded by spaces",
                 CodeModificationHint());
    }
  }
}

void RuleChecker::checkParen(ParenExpr* E) {
  SourceLocation LS = E->getLParen();
  if (!LS.isValid() || !LS.isFileID()) return;

  SourceLocation LE = E->getRParen();
  if (!LE.isValid() || !LE.isFileID()) return;

  unsigned ColNoS = 0, ColNoE = 0;
  const char* TokPtr = getTokenPtr(LS, ColNoS);
  const char* TokEnd = getTokenPtr(LE, ColNoE);

  if (ColNoS > 1 && TokEnd - TokPtr > 1) {
    if (isspace(*(TokPtr+1)) || isspace(*(TokEnd-1))) {
      CCC.report(E, "Opening [closing] parentheses must not be followed "
                 "[preceded] by spaces",
                 CodeModificationHint());
    }
  }
}

bool RuleChecker::processConfigLine(llvm::StringRef Line, size_t LineNo) {
  std::string SKind;
  int SValue;

  std::istringstream InS(Line);

  SKind = "<NOT READ>";
  InS >> SKind;

  if (InS) {
    InS >> SValue;
  }
  if (!InS) {
    std::stringstream SErr;
    if (SKind == "<NOT READ>") {
      SErr << "rule's type of check.";
    } else {
      SErr << "rule's value.";
    }
    CCC.reportConfigError(SErr.str(), LineNo);
    Rules.clear();
    return false;
  }
    
  Rules[SKind] = SValue;
  return true;
}


void clang::CheckCodingConventionSpaces(Decl *D, Diagnostic& Diag,
                                        SourceManager& SM,
                                        const LangOptions& Lang) {
  CodingConventionChecker CCC(Diag, SM, Lang);
  RuleChecker Checker(CCC);
  CCC.readConfig("Spaces", Checker);
  WalkAST walker(Checker);
  walker.Visit(D);
}

